package java.util.stream;

import java.util.Spliterator;
import java.util.function.Supplier;
import java.util.function.DoublePredicate;
import java.util.function.LongPredicate;
import java.util.function.IntPredicate;
import java.util.Objects;
import java.util.function.Predicate;

final class MatchOps
{
    private MatchOps() {
    }
    
    public static <T> TerminalOp<T, Boolean> makeRef(Predicate<? super T> obj, MatchOps.MatchKind obj2) {
        Objects.requireNonNull(obj);
        Objects.requireNonNull(obj2);
        class MatchSink extends MatchOps.BooleanTerminalSink<T>
        {
            MatchSink() {
                super(matchKind);
            }
            
            @Override
            public void accept(T t) {
                if (!this.stop && p0.test(t) == p1.stopOnPredicateMatches) {
                    this.stop = true;
                    this.value = p1.shortCircuitResult;
                }
            }
        }
        return new MatchOps.MatchOp(StreamShape.REFERENCE, obj2, ()/*  */ -> new MatchSink());
    }
    
    public static TerminalOp<Integer, Boolean> makeInt(IntPredicate obj, MatchOps.MatchKind obj2) {
        Objects.requireNonNull(obj);
        Objects.requireNonNull(obj2);
        class MatchSink extends MatchOps.BooleanTerminalSink<Integer> implements Sink.OfInt
        {
            MatchSink() {
                super(matchKind);
            }
            
            @Override
            public void accept(int n) {
                if (!this.stop && p0.test(n) == p1.stopOnPredicateMatches) {
                    this.stop = true;
                    this.value = p1.shortCircuitResult;
                }
            }
        }
        return new MatchOps.MatchOp(StreamShape.INT_VALUE, obj2, ()/*  */ -> new MatchSink());
    }
    
    public static TerminalOp<Long, Boolean> makeLong(LongPredicate obj, MatchOps.MatchKind obj2) {
        Objects.requireNonNull(obj);
        Objects.requireNonNull(obj2);
        class MatchSink extends MatchOps.BooleanTerminalSink<Long> implements Sink.OfLong
        {
            MatchSink() {
                super(matchKind);
            }
            
            @Override
            public void accept(long n) {
                if (!this.stop && p0.test(n) == p1.stopOnPredicateMatches) {
                    this.stop = true;
                    this.value = p1.shortCircuitResult;
                }
            }
        }
        return new MatchOps.MatchOp(StreamShape.LONG_VALUE, obj2, ()/*  */ -> new MatchSink());
    }
    
    public static TerminalOp<Double, Boolean> makeDouble(DoublePredicate obj, MatchOps.MatchKind obj2) {
        Objects.requireNonNull(obj);
        Objects.requireNonNull(obj2);
        class MatchSink extends MatchOps.BooleanTerminalSink<Double> implements Sink.OfDouble
        {
            MatchSink() {
                super(matchKind);
            }
            
            @Override
            public void accept(double n) {
                if (!this.stop && p0.test(n) == p1.stopOnPredicateMatches) {
                    this.stop = true;
                    this.value = p1.shortCircuitResult;
                }
            }
        }
        return new MatchOps.MatchOp(StreamShape.DOUBLE_VALUE, obj2, ()/*  */ -> new MatchSink());
    }
    
    enum MatchKind
    {
        ANY(true, true), 
        ALL(false, false), 
        NONE(true, false);
        
        private final boolean stopOnPredicateMatches;
        private final boolean shortCircuitResult;
        
        private MatchKind(boolean stopOnPredicateMatches, boolean shortCircuitResult) {
            this.stopOnPredicateMatches = stopOnPredicateMatches;
            this.shortCircuitResult = shortCircuitResult;
        }
    }
    
    private static final class MatchOp<T> implements TerminalOp<T, Boolean>
    {
        private final StreamShape inputShape;
        final MatchOps.MatchKind matchKind;
        final Supplier<MatchOps.BooleanTerminalSink<T>> sinkSupplier;
        
        MatchOp(StreamShape inputShape, MatchOps.MatchKind matchKind, Supplier<MatchOps.BooleanTerminalSink<T>> sinkSupplier) {
            this.inputShape = inputShape;
            this.matchKind = matchKind;
            this.sinkSupplier = sinkSupplier;
        }
        
        @Override
        public int getOpFlags() {
            return StreamOpFlag.IS_SHORT_CIRCUIT | StreamOpFlag.NOT_ORDERED;
        }
        
        @Override
        public StreamShape inputShape() {
            return this.inputShape;
        }
        
        @Override
        public <S> Boolean evaluateSequential(PipelineHelper<T> pipelineHelper, Spliterator<S> spliterator) {
            return Boolean.valueOf(((MatchOps.BooleanTerminalSink)pipelineHelper.wrapAndCopyInto((Sink)this.sinkSupplier.get(), spliterator)).getAndClearState());
        }
        
        @Override
        public <S> Boolean evaluateParallel(PipelineHelper<T> pipelineHelper, Spliterator<S> spliterator) {
            return (Boolean)new MatchOps.MatchTask(this, pipelineHelper, spliterator).invoke();
        }
    }
    
    private abstract static class BooleanTerminalSink<T> implements Sink<T>
    {
        boolean stop;
        boolean value;
        
        BooleanTerminalSink(MatchOps.MatchKind matchKind) {
            this.value = !matchKind.shortCircuitResult;
        }
        
        public boolean getAndClearState() {
            return this.value;
        }
        
        @Override
        public boolean cancellationRequested() {
            return this.stop;
        }
    }
    
    private static final class MatchTask<P_IN, P_OUT> extends AbstractShortCircuitTask<P_IN, P_OUT, Boolean, MatchOps.MatchTask<P_IN, P_OUT>>
    {
        private final MatchOps.MatchOp<P_OUT> op;
        
        MatchTask(MatchOps.MatchOp<P_OUT> op, PipelineHelper<P_OUT> pipelineHelper, Spliterator<P_IN> spliterator) {
            super(pipelineHelper, spliterator);
            this.op = op;
        }
        
        MatchTask(MatchOps.MatchTask<P_IN, P_OUT> matchTask, Spliterator<P_IN> spliterator) {
            super(matchTask, spliterator);
            this.op = matchTask.op;
        }
        
        @Override
        protected MatchOps.MatchTask<P_IN, P_OUT> makeChild(Spliterator<P_IN> spliterator) {
            return new MatchOps.MatchTask<P_IN, P_OUT>(this, spliterator);
        }
        
        @Override
        protected Boolean doLeaf() {
            int andClearState = ((MatchOps.BooleanTerminalSink)this.helper.wrapAndCopyInto((Sink)this.op.sinkSupplier.get(), this.spliterator)).getAndClearState();
            if (andClearState == (this.op.matchKind.shortCircuitResult ? 1 : 0)) {
                this.shortCircuit(Boolean.valueOf((boolean)andClearState));
            }
            return null;
        }
        
        @Override
        protected Boolean getEmptyResult() {
            return Boolean.valueOf(!this.op.matchKind.shortCircuitResult);
        }
    }
}
